#!/usr/bin/python3

# ostree-push - Push local ostree repo to remote server over SSH
# Copyright (C) 2017  Endless Mobile, Inc.
# Copyright (C) 2021  Endless OS Foundation LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Push local ostree repo to remote server over SSH

ostree-push allows coherently publishing an ostree repo to a remote
server. It works by opening an SSH connection to the remote server and
initiating the ostree-receive service to pull from an HTTP server on the
local host. This has a distinct advantage over using rsync where files
can be pushed in the wrong order and there's no ability to push a subset
of the refs in the local repo.

ostree-push will start an HTTP server and tunnel its port to the remote
server. This allows publishing from a host that is not running an HTTP
server and avoids any firewalls between the local and remote hosts.

In either case, ostree-receive must be installed on the remote host to
pull from the tunnelled HTTP server.
"""

from . import VERSION

from argparse import Action, ArgumentError, ArgumentParser
from collections import namedtuple
import gi
from http.server import SimpleHTTPRequestHandler
import logging
import multiprocessing
import os
import queue
import shlex
import subprocess
from tempfile import TemporaryDirectory
import threading
import time
from urllib.parse import urlparse

try:
    from http.server import ThreadingHTTPServer
except ImportError:
    from http.server import HTTPServer
    from socketserver import ThreadingMixIn

    class ThreadingHTTPServer(ThreadingMixIn, HTTPServer):
        daemon_threads = True

gi.require_version('OSTree', '1.0')
from gi.repository import Gio, OSTree  # noqa: E402

logger = logging.getLogger(__name__)

# Timeout in seconds when waiting for ports or sockets
RESOURCE_TIMEOUT = 60


class OTPushError(Exception):
    """Exceptions from ostree-push"""
    pass


class OSTreeRequestHandler(SimpleHTTPRequestHandler):
    """SimpleHTTPRequestHandler with logging"""
    def log_message(self, format, *args):
        logger.debug("%s: %s - - [%s] %s",
                     threading.current_thread().name,
                     self.address_string(),
                     self.log_date_time_string(),
                     format % args)


class RepoServer:
    """HTTP server for repo

    Start an HTTP server running at the repository path. The server
    listens on an ephemeral port on the loopback address. The timeout
    parameter controls how long to wait for the server process to send
    back its URL.

    When used as a context manager, the server is stopped when the
    context closes.
    """
    def __init__(self, path, timeout=RESOURCE_TIMEOUT):
        self.path = path
        self.timeout = timeout
        self.proc = None
        self.address = None
        self.url = None

        if not os.path.isdir(path):
            raise ValueError('{} is not a directory'.format(path))

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()

    def _run_server(self, path, queue):
        # FIXME: When python 3.7 is the minimum, use the
        # SimpleHTTPRequestHandler directory parameter with
        # functools.partial instead of changing directory.
        os.chdir(path)
        server = ThreadingHTTPServer(('127.0.0.1', 0), OSTreeRequestHandler)
        queue.put(server.server_address)
        server.serve_forever()

    def start(self):
        """Start an HTTP server for the repo

        The server is started in a separate process which send back the
        address it is bound to. If the address is not received by the
        instance's timeout value, an exception is raised.
        """
        addr_queue = multiprocessing.Queue()
        self.proc = multiprocessing.Process(
            target=self._run_server,
            args=(self.path, addr_queue)
        )
        self.proc.start()
        if not self.proc.is_alive():
            raise OTPushError(f'HTTP process {self.proc.pid} exited')
        try:
            self.address = addr_queue.get(True, self.timeout)
        except queue.Empty:
            raise OTPushError('HTTP process did not write port to queue') \
                from None

        self.url = f'http://{self.address[0]}:{self.address[1]}'
        logger.info('Serving %s on %s from process %d', self.path, self.url,
                    self.proc.pid)

    def stop(self):
        """Stop the HTTP server"""
        if self.proc is not None:
            if self.proc.is_alive():
                logger.debug('Stopping HTTP server process %d', self.proc.pid)
                self.proc.terminate()
            self.proc = None


class SSHMultiplexer:
    """SSH multiplexer for connecting with a remote host

    The remote host and a path to a non-existent socket.
    """
    def __init__(self, host, socket, ssh_options=None, user=None,
                 port=None):
        self.host = host
        self.user = user
        self.port = port
        self.socket = socket
        self.ssh_options = ssh_options
        self.master_proc = None

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()

    def __del__(self):
        self.stop()

    def start(self):
        """Start an SSH master connection

        Run an SSH master connection to host in the background.
        """
        if self.master_proc is not None:
            raise OTPushError(
                f'SSH master process already running in {self.master_proc.pid}'
            )
        if os.path.exists(self.socket):
            raise OTPushError(f'Socket {self.socket} already exists')

        # Create the socket file if necessary
        # Options used:
        # -N: Don't execute a remote command
        # -M: Puts the client in master mode for connection sharing
        # -S: Specify the location of the control socket
        master_cmd = ['ssh', '-N', '-M', '-S', self.socket]
        if self.port:
            master_cmd += ['-p', str(self.port)]
        if self.ssh_options:
            master_cmd += self.ssh_options
        if self.user:
            master_cmd.append(f'{self.user}@{self.host}')
        else:
            master_cmd.append(self.host)
        logger.debug('Starting SSH master process %s',
                     ' '.join(map(shlex.quote, master_cmd)))
        self.master_proc = subprocess.Popen(master_cmd)

        # Loop until the socket shows up
        timeout = 0
        while timeout < RESOURCE_TIMEOUT:
            if self.master_proc.poll() is not None:
                raise OTPushError(
                    f'SSH master process {self.master_proc.pid} exited')
            if os.path.exists(self.socket):
                break
            timeout += 0.5
            time.sleep(0.5)

        if timeout >= RESOURCE_TIMEOUT:
            raise OTPushError(
                f'SSH control socket {self.socket} does not exist')

    def stop(self):
        if self.master_proc is not None:
            if self.master_proc.poll() is None:
                logger.debug('Stopping SSH master process %d',
                             self.master_proc.pid)
                self.master_proc.terminate()
            self.master_proc = None

    def forward_port(self, port):
        """Forward a local port over an SSH master connection

        Forward a local port to host over the SSH master connection. The
        remote port is returned.
        """
        if self.master_proc is None:
            raise OTPushError('SSH master process not running')

        # Options used:
        # -S: Specify the location of the control socket
        # -O: Makes the client print the forwarded port on stdout
        # -R: Forward the local port to the remote. Use 0 as the remote port
        #     so it binds one dynamically.
        forward_cmd = [
            'ssh', '-S', self.socket, '-O', 'forward',
            '-R', f'0:127.0.0.1:{port}'
        ]
        if self.ssh_options:
            forward_cmd += self.ssh_options
        forward_cmd.append(self.host)
        logger.debug('Forwarding HTTP port with %s',
                     ' '.join(map(shlex.quote, forward_cmd)))
        out = subprocess.check_output(forward_cmd)

        # Pass the output through int() so that any whitespace is stripped
        # and we know a port number was returned. Anything else will raise a
        # ValueError.
        return int(out)

    def run(self, cmd):
        """Run a command on the remote host using the master connection"""
        if self.master_proc is None:
            raise OTPushError('SSH master process not running')

        run_cmd = ['ssh', '-S', self.socket]
        if self.ssh_options:
            run_cmd += self.ssh_options
        run_cmd += [self.host] + cmd
        logger.debug('Executing ' + ' '.join(map(shlex.quote, cmd)))
        subprocess.check_call(run_cmd)


def push_refs(local_repo, dest, refs=None, ssh_options=None,
              command='ostree-receive', dry_run=False):
    """Run ostree-receive on remote with a tunneled HTTP server

    Start a local HTTP server and tunnel its port to the remote host.
    Use this tunneled HTTP server as the URL for ostree_receive().
    """
    local_repo_path = local_repo.get_path().get_path()

    # If refs were specified, make sure they exist before bothering with
    # the remote connection
    if refs:
        list_refs_flags = OSTree.RepoListRefsExtFlags.EXCLUDE_REMOTES
        try:
            # EXCLUDE_MIRRORS only available since ostree 2019.2
            list_refs_flags |= OSTree.RepoListRefsExtFlags.EXCLUDE_MIRRORS
        except AttributeError:
            pass
        _, local_refs = local_repo.list_refs_ext(None, list_refs_flags)
        missing_refs = sorted(set(refs) - local_refs.keys())
        if missing_refs:
            raise OTPushError(
                f'Refs {" ".join(missing_refs)} not found in {local_repo_path}'
            )

    summary = os.path.join(local_repo_path, 'summary')
    update_summary = False
    if not os.path.exists(summary):
        logger.debug('%s does not exist, regenerating', summary)
        update_summary = True
    else:
        # OSTree updates the mtime of the repo directory when refs have
        # been updated, so if that's newer than the summary, it needs to
        # be regenerated.
        repo_mtime = os.path.getmtime(local_repo_path)
        summary_mtime = os.path.getmtime(summary)
        if summary_mtime < repo_mtime:
            logger.debug('Repo %s has been modified more recently than %s, '
                         'regenerating',
                         local_repo_path, summary)
            update_summary = True
    if update_summary:
        logger.info('Regenerating summary file')
        local_repo.regenerate_summary()

    with RepoServer(local_repo_path) as http_server:
        http_port = http_server.address[1]

        with TemporaryDirectory(prefix='ostree-push-') as tempdir:
            socket_path = os.path.join(tempdir, 'socket')

            # Start an SSH master, forward the HTTP port to the remote
            # server and run ostree-receive there
            with SSHMultiplexer(dest.host, socket_path, ssh_options,
                                user=dest.user, port=dest.port) as ssh:
                remote_port = ssh.forward_port(http_port)
                logger.info('Connected local HTTP port %d to remote port %d',
                            http_port, remote_port)
                remote_url = f'http://127.0.0.1:{remote_port}'

                cmd = shlex.split(command)
                if dry_run:
                    cmd.append('-n')
                cmd += [dest.repo, remote_url]
                if refs is not None:
                    cmd += refs
                logger.debug('Remote command: %s', cmd)
                ssh.run(cmd)


PushDest = namedtuple('PushDest', ('host', 'repo', 'user', 'port'))


def parse_dest(dest):
    """Parse the push destination into host and repo

    Allowed destinations are:

    * [user@]host:path/to/repo
    * ssh://[user@]host[:port]/path/to/repo
    """
    # A bit of care is needed because urlparse parses host:repo into a
    # scheme and path but user@host:repo into just a path.
    parts = urlparse(dest)
    if parts.scheme and parts.netloc:
        # ssh:// URL
        if parts.scheme != 'ssh':
            raise ValueError(
                f'Destination scheme "{parts.scheme}" not allowed')
        if not parts.path:
            raise ValueError('Destintion repo missing')

        return PushDest(host=parts.hostname, user=parts.username,
                        port=parts.port, repo=parts.path)
    else:
        # scp form. There should be at least 1 : to separate the host
        # and repo.
        host, sep, repo = dest.partition(':')
        if not sep:
            raise ValueError('Destination not in form "host:repo"')
        if not host:
            raise ValueError('Destination host missing')
        if not repo:
            raise ValueError('Destination repo missing')

        # Try to split user@ from the host.
        user = None
        tmpuser, sep, tmphost = host.partition('@')
        if sep:
            if not tmpuser or not tmphost:
                raise ValueError(f'Invalid destination host {host}')
            host = tmphost
            user = tmpuser

        return PushDest(host=host, user=user, repo=repo, port=None)


class DestArgAction(Action):
    """Action to set push destination"""
    def __init__(self, option_strings, dest, nargs=None, default=None,
                 **kwargs):
        if nargs is not None:
            raise ValueError('nargs not allowed')
        if default is not None:
            raise ValueError('default not allowed')

        super().__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string):
        dest = parse_dest(values)
        setattr(namespace, self.dest, dest)


class SSHOptAction(Action):
    """Action to collect ssh option verbatim"""
    # Options that can be specified multiple times
    MULTI_OPTS = {'-o'}

    def __init__(self, option_strings, dest, nargs=None, default=None,
                 **kwargs):
        if option_strings is None:
            raise ValueError('option strings are required')
        if nargs is not None:
            raise ValueError('nargs not allowed')
        if default is not None:
            raise ValueError('default not allowed')

        dest = 'ssh_options'
        default = []
        super().__init__(option_strings, dest, default=default, **kwargs)

    def __call__(self, parser, namespace, values, option_string):
        logger.debug('Parsing SSH option %r, %r, %r', namespace, option_string,
                     values)

        # A bit of care is needed in case parse_args() is called
        # multiple times. The default list needs to be copied so it's
        # not updated and the set of options seen needs to be reset.
        ssh_opts = getattr(namespace, self.dest)
        if ssh_opts is self.default:
            ssh_opts = ssh_opts.copy()
            self._single_opts_seen = set()

        if option_string not in self.MULTI_OPTS:
            if option_string in self._single_opts_seen:
                raise ArgumentError(
                    self,
                    f'Option {option_string} can only be specified once'
                )
            self._single_opts_seen.add(option_string)

        ssh_opts += [option_string, values]
        setattr(namespace, self.dest, ssh_opts)


class OTPushArgParser(ArgumentParser):
    """ArgumentParser for ostree-push"""
    def __init__(self):
        super().__init__(
            description='Push ostree refs to remote repository'
        )

        self.add_argument(
            'dest', metavar='DEST', action=DestArgAction,
            help=(
                'remote host and repo destination. DEST can take one of two '
                'forms: [user@]host:path/to/repo '
                'or ssh://[user@]host[:port]/path/to/repo.'
            )
        )
        self.add_argument('refs', metavar='REF', nargs='*', default=[],
                          help='ostree refs to push, all if none specified')
        self.add_argument('-n', '--dry-run', action='store_true',
                          help='only show what would be done')
        self.add_argument('-p', dest='port', type=int,
                          help='ssh port to connect to')
        self.set_defaults(log_level=logging.INFO)
        self.add_argument('-q', '--quiet', dest='log_level',
                          action='store_const', const=logging.WARNING,
                          help='disable most messages')
        self.add_argument('-v', '--verbose', dest='log_level',
                          action='store_const', const=logging.DEBUG,
                          help='enable verbose messages')
        self.add_argument(
            '--repo',
            help='local repository path (default: current directory)'
        )
        self.add_argument('--command', default='ostree-receive',
                          help='remote pull command (default: %(default)s)')
        self.add_argument('-i', '-o', metavar='OPTION',
                          action=SSHOptAction,
                          help='options passed through to ssh')
        self.add_argument('--version', action='version',
                          version=f'%(prog)s {VERSION}')

    def parse_args(self, *args, **kwargs):
        args = super().parse_args(*args, **kwargs)

        # If a port option has been supplied, replace the port in the
        # dest argument with it.
        if args.port:
            args.dest = args.dest._replace(port=args.port)

        return args


def main(argv=None):
    aparser = OTPushArgParser()
    args = aparser.parse_args(argv)

    logging.basicConfig(level=args.log_level)

    # Open the local repo and get the full path
    if args.repo:
        repo_file = Gio.File.new_for_path(args.repo)
        repo = OSTree.Repo.new(repo_file)
    else:
        repo = OSTree.Repo.new_default()
    repo.open()

    push_refs(repo, args.dest, refs=args.refs,
              ssh_options=args.ssh_options, command=args.command,
              dry_run=args.dry_run)


if __name__ == '__main__':
    main()



import json
import os
import pytest
import shutil
import subprocess

from .util import SCRIPTSDIR, TESTSDIR

shell_abspath = os.path.join(SCRIPTSDIR, 'ostree-receive-shell')
dumpenv_abspath = os.path.join(TESTSDIR, 'dumpenv')

# Some tests can't be run if ostree-receive is in PATH
skip_ostree_receive_in_path = pytest.mark.skipif(
    shutil.which('ostree-receive') is not None,
    reason="cannot test correctly with ostree-receive in PATH"
)


@pytest.fixture
def tmp_bindir(tmp_path):
    """A temporary directory to be included in PATH"""
    bindir = tmp_path / 'bin'
    bindir.mkdir()
    return bindir


@pytest.fixture
def tmp_receive(tmp_bindir):
    """Copy dumpenv to a temporary ostree-receive"""
    receive = tmp_bindir / 'ostree-receive'
    shutil.copy(dumpenv_abspath, receive)
    os.chmod(receive, 0o755)
    return receive


@pytest.fixture
def tmp_shell(tmp_bindir):
    """Copy ostree-receive-shell to the temporary bindir"""
    shell = tmp_bindir / 'ostree-receive-shell'
    shutil.copy(shell_abspath, shell)
    os.chmod(shell, 0o755)
    return shell


@pytest.fixture
def shell_env_vars(tmp_bindir):
    """Environment variables for shell tests"""
    env = os.environ.copy()
    path = env.get('PATH', os.defpath).split(os.pathsep)
    path.insert(0, str(tmp_bindir))
    env['PATH'] = os.pathsep.join(path)
    return env


def test_command_args(shell_env_vars, tmp_shell, tmp_receive):
    """Test how arguments are passed to ostree-receive"""
    cmd = ('ostree-receive-shell', '-c', 'ostree-receive')
    proc = subprocess.run(cmd, check=True, env=shell_env_vars,
                          stdout=subprocess.PIPE)
    data = json.loads(proc.stdout.decode('utf-8'))
    assert data['args'] == [str(tmp_receive)]

    cmd = ('ostree-receive-shell', '-c', 'ostree-receive -n foo bar')
    proc = subprocess.run(cmd, check=True, env=shell_env_vars,
                          stdout=subprocess.PIPE)
    data = json.loads(proc.stdout.decode('utf-8'))
    assert data['args'] == [str(tmp_receive), '-n', 'foo', 'bar']


def test_auto_path(shell_env_vars, tmp_receive):
    """Test that the shell's directory is appended to PATH"""
    # Here we use the shell in the source directory to ensure that it's
    # directory isn't in PATH. Otherwise it won't get appended.
    cmd = (shell_abspath, '-c', 'ostree-receive')
    proc = subprocess.run(cmd, check=True, env=shell_env_vars,
                          stdout=subprocess.PIPE)
    data = json.loads(proc.stdout.decode('utf-8'))
    path = data['env']['PATH'].split(os.pathsep)
    assert path[-1] == SCRIPTSDIR


def test_no_interactive():
    """Test trying to run the shell interactively with no arguments"""
    cmd = (shell_abspath,)
    proc = subprocess.run(cmd, stderr=subprocess.PIPE)
    assert proc.returncode == 1
    assert proc.stderr.decode('utf-8') == (
        'ostree-receive-shell: Cannot run interactively\n'
    )


def test_wrong_args():
    """Test passing incorrect arguments"""
    commands = (
        (shell_abspath, 'foo'),
        (shell_abspath, 'foo', 'bar'),
        (shell_abspath, 'foo', 'bar', 'baz'),
        (shell_abspath, '-c'),
        (shell_abspath, '-c', 'foo', 'bar'),
    )
    for cmd in commands:
        proc = subprocess.run(cmd, stderr=subprocess.PIPE)
        assert proc.returncode == 1
        assert proc.stderr.decode('utf-8') == (
            'ostree-receive-shell: Must be run with no arguments or '
            'with -c cmd\n'
        )


def test_bad_command():
    """Test when disallowed commands are requested"""
    arguments = (
        ('foo',),
        ('foo', 'bar'),
        ('/usr/bin/ostree-receive'),
    )
    for args in arguments:
        cmd = (shell_abspath, '-c', ' '.join(args))
        proc = subprocess.run(cmd, stderr=subprocess.PIPE)
        assert proc.returncode == 1
        assert proc.stderr.decode('utf-8') == (
            f'ostree-receive-shell: Executing {args[0]} not allowed\n'
        )


# This test depends on the temporary ostree-receive being the only one
# in PATH, so it has to be skipped otherwise. It might be possible to
# mangle PATH so it's avoided, but then the required python might also
# be removed from PATH.
@skip_ostree_receive_in_path
def test_exec_errors(shell_env_vars, tmp_shell, tmp_receive, tmp_path):
    """Test how errors from execve are handled"""
    cmd = ('ostree-receive-shell', '-c', 'ostree-receive')

    # Make the temporary ostree-receive non-executable to get a
    # permission denied error.
    tmp_receive.chmod(0o644)
    proc = subprocess.run(cmd, env=shell_env_vars, stderr=subprocess.PIPE)
    assert proc.returncode == 126
    assert proc.stderr.decode('utf-8') == (
        'ostree-receive-shell: ostree-receive: Permission denied\n'
    )

    # Make the temporary ostree-receive into a dangling symlink to get a
    # file not found error.
    tmp_receive.unlink()
    tmp_receive.symlink_to(tmp_path / 'nonexistent')
    proc = subprocess.run(cmd, env=shell_env_vars, stderr=subprocess.PIPE)
    assert proc.returncode == 127
    assert proc.stderr.decode('utf-8') == (
        'ostree-receive-shell: ostree-receive: No such file or directory\n'
    )



from otpush import receive

import argparse
import dataclasses
import gi
from gi.repository import GLib, Gio
import json
import logging
import os
from pathlib import Path
import pytest
import time
import yaml

from .util import (
    ED25519_PRIVATE_KEY,
    ED25519_PUBLIC_KEY,
    PGP_PUB,
    PGP_PUB_KEYRING,
    PGP_KEY_ID,
    TESTSDIR,
    get_ostree_ed25519_sign,
    get_summary_variant,
    local_refs,
    needs_ed25519,
    needs_flatpak,
    needs_gpg,
    needs_ostree,
    oneshot_transaction,
    random_commit,
    wipe_repo,
    TmpRepo,
)

gi.require_version('OSTree', '1.0')
from gi.repository import OSTree  # noqa: E402

logger = logging.getLogger(__name__)


class TestReceiveRepo:
    def test_cleanup(self, dest_repo):
        url = 'http://example.com'
        repo = receive.OTReceiveRepo(dest_repo.path, url)
        remotes_dir = Path(repo.remotes_dir.name)
        assert remotes_dir.exists()
        del repo
        assert not remotes_dir.exists()

        with receive.OTReceiveRepo(dest_repo.path, url) as repo:
            remotes_dir = Path(repo.remotes_dir.name)
            assert remotes_dir.exists()
        assert not remotes_dir.exists()

    def test_get_commit_timestamp(self, tmp_files_path, receive_repo):
        with pytest.raises(GLib.Error) as excinfo:
            receive_repo._get_commit_timestamp('missing')
        assert excinfo.value.matches(Gio.io_error_quark(),
                                     Gio.IOErrorEnum.NOT_FOUND)

        commit = random_commit(receive_repo, tmp_files_path, 'someref',
                               timestamp=0)
        timestamp = receive_repo._get_commit_timestamp(commit)
        assert timestamp == 0

        now = int(time.time())
        commit = random_commit(receive_repo, tmp_files_path, 'someref',
                               timestamp=now)
        timestamp = receive_repo._get_commit_timestamp(commit)
        assert timestamp == now

        now = int(time.time())
        commit = random_commit(receive_repo, tmp_files_path, 'someref')
        timestamp = receive_repo._get_commit_timestamp(commit)
        assert timestamp >= now

    def test_is_flatpak_repo(self, tmp_files_path, receive_repo):
        assert not receive_repo._is_flatpak_repo()

        random_commit(receive_repo, tmp_files_path, 'someref')
        assert not receive_repo._is_flatpak_repo()

        random_commit(receive_repo, tmp_files_path,
                      'app/com.example.App/x86_64/stable')
        assert receive_repo._is_flatpak_repo()

    def test_pull_commits(self, tmp_files_path, receive_repo, source_repo,
                          source_server):
        rev1 = random_commit(source_repo, tmp_files_path, 'ref1')
        rev2 = random_commit(source_repo, tmp_files_path, 'ref2')

        _, remote_refs = receive_repo.remote_list_refs(
            receive_repo.REMOTE_NAME)
        assert remote_refs == {'ref1': rev1, 'ref2': rev2}

        # Pull by single ref
        with oneshot_transaction(receive_repo):
            receive_repo._pull_commits(['ref1'])
        _, refs = receive_repo.list_refs(None)
        assert refs == {'_receive:ref1': rev1}
        _, _, state = receive_repo.load_commit(rev1)
        assert state == OSTree.RepoCommitState.NORMAL
        wipe_repo(receive_repo)

        # Pull by multiple refs
        with oneshot_transaction(receive_repo):
            receive_repo._pull_commits(['ref1', 'ref2'])
        _, refs = receive_repo.list_refs(None)
        assert refs == {'_receive:ref1': rev1, '_receive:ref2': rev2}
        _, _, state = receive_repo.load_commit(rev1)
        assert state == OSTree.RepoCommitState.NORMAL
        _, _, state = receive_repo.load_commit(rev2)
        assert state == OSTree.RepoCommitState.NORMAL
        wipe_repo(receive_repo)

        # Pull by single rev
        with oneshot_transaction(receive_repo):
            receive_repo._pull_commits([rev1])
        _, refs = receive_repo.list_refs(None)
        assert refs == {}
        _, _, state = receive_repo.load_commit(rev1)
        assert state == OSTree.RepoCommitState.NORMAL
        wipe_repo(receive_repo)

        # Pull by multiple revs
        with oneshot_transaction(receive_repo):
            receive_repo._pull_commits([rev1, rev2])
        _, refs = receive_repo.list_refs(None)
        assert refs == {}
        _, _, state = receive_repo.load_commit(rev1)
        assert state == OSTree.RepoCommitState.NORMAL
        _, _, state = receive_repo.load_commit(rev2)
        assert state == OSTree.RepoCommitState.NORMAL
        wipe_repo(receive_repo)

        # Pull by missing ref
        with pytest.raises(GLib.Error) as excinfo:
            with oneshot_transaction(receive_repo):
                receive_repo._pull_commits(['missing'])
        assert excinfo.value.matches(Gio.io_error_quark(),
                                     Gio.IOErrorEnum.FAILED)
        wipe_repo(receive_repo)

    def test_copy_commit(self, tmp_files_path, receive_repo):
        # Non-flatpak ref
        src = random_commit(receive_repo, tmp_files_path, 'src')
        with oneshot_transaction(receive_repo):
            dst = receive_repo.copy_commit(src, 'dst')
        assert src != dst
        _, src_commit, _ = receive_repo.load_commit(src)
        _, dst_commit, dst_state = receive_repo.load_commit(dst)
        assert dst_state == OSTree.RepoCommitState.NORMAL
        assert OSTree.commit_get_parent(dst_commit) is None
        assert OSTree.commit_get_content_checksum(src_commit) == \
            OSTree.commit_get_content_checksum(dst_commit)
        assert OSTree.commit_get_timestamp(src_commit) == \
            OSTree.commit_get_timestamp(dst_commit)
        expected_metadata = {
            OSTree.COMMIT_META_KEY_REF_BINDING: ['dst'],
            OSTree.COMMIT_META_KEY_COLLECTION_BINDING: 'com.example.Test',
        }
        dst_metadata = dst_commit.get_child_value(0).unpack()
        assert dst_metadata == expected_metadata

        # Make another commit and check that the previous commit is used
        # as the parent
        expected_parent = dst
        src = random_commit(receive_repo, tmp_files_path, 'src')
        with oneshot_transaction(receive_repo):
            dst = receive_repo.copy_commit(src, 'dst')
        assert src != dst
        _, src_commit, _ = receive_repo.load_commit(src)
        _, dst_commit, dst_state = receive_repo.load_commit(dst)
        assert dst_state == OSTree.RepoCommitState.NORMAL
        assert OSTree.commit_get_parent(dst_commit) == expected_parent
        assert OSTree.commit_get_content_checksum(src_commit) == \
            OSTree.commit_get_content_checksum(dst_commit)
        assert OSTree.commit_get_timestamp(src_commit) == \
            OSTree.commit_get_timestamp(dst_commit)
        dst_metadata = dst_commit.get_child_value(0).unpack()
        assert dst_metadata == expected_metadata

        # Check that additional metadata is copied
        extra_metadata = {'foo': GLib.Variant('s', 'bar')}
        expected_metadata['foo'] = 'bar'
        src = random_commit(receive_repo, tmp_files_path, 'src',
                            extra_metadata=extra_metadata)
        with oneshot_transaction(receive_repo):
            dst = receive_repo.copy_commit(src, 'dst')
        assert src != dst
        _, dst_commit, _ = receive_repo.load_commit(dst)
        dst_metadata = dst_commit.get_child_value(0).unpack()
        assert dst_metadata == expected_metadata

        # Flatpak metadata
        ref = 'app/com.example.Foo/x86_64/stable'
        src = random_commit(receive_repo, tmp_files_path, 'src')
        with oneshot_transaction(receive_repo):
            dst = receive_repo.copy_commit(src, ref)
        assert src != dst
        _, dst_commit, _ = receive_repo.load_commit(dst)
        expected_metadata = {
            OSTree.COMMIT_META_KEY_REF_BINDING: [ref],
            OSTree.COMMIT_META_KEY_COLLECTION_BINDING: 'com.example.Test',
            'xa.ref': ref,
            'xa.from_commit': src,
        }
        dst_metadata = dst_commit.get_child_value(0).unpack()
        assert dst_metadata == expected_metadata

        # Copying partial commit should fail
        receive_repo.mark_commit_partial(src, True)
        with pytest.raises(receive.OTReceiveError) as excinfo:
            with oneshot_transaction(receive_repo):
                receive_repo.copy_commit(src, ref)
        assert str(excinfo.value) == f'Cannot copy irregular commit {src}'

    def test_receive(self, tmp_files_path, receive_repo, source_repo,
                     source_server):
        random_commit(source_repo, tmp_files_path, 'ref1')
        random_commit(source_repo, tmp_files_path, 'ref2')

        with pytest.raises(receive.OTReceiveError) as excinfo:
            receive_repo.receive(['missing'])
        assert str(excinfo.value) == \
            'Could not find ref missing in summary file'

        merged = receive_repo.receive(['ref1'])
        assert merged == {'ref1'}
        refs = local_refs(receive_repo)
        assert refs.keys() == {'ref1'}

    @needs_ostree
    def test_receive_update(self, tmp_files_path, receive_repo, source_repo,
                            source_server):
        receive_repo.config.update = True

        random_commit(source_repo, tmp_files_path, 'ref1')

        merged = receive_repo.receive(['ref1'])
        assert merged == {'ref1'}
        refs = local_refs(receive_repo)
        assert refs.keys() == {'ostree-metadata', 'ref1'}
        summary = Path(receive_repo.path) / 'summary'
        assert summary.exists()

    def test_receive_update_hook(self, tmp_files_path, receive_repo,
                                 source_repo, source_server):
        dumpenv = os.path.join(TESTSDIR, 'dumpenv')
        receive_repo.config.update = True
        receive_repo.config.update_hook = dumpenv

        random_commit(source_repo, tmp_files_path, 'ref1')

        merged = receive_repo.receive(['ref1'])
        assert merged == {'ref1'}
        refs = local_refs(receive_repo)
        assert refs.keys() == {'ref1'}
        summary = Path(receive_repo.path) / 'summary'
        assert not summary.exists()

    @needs_gpg
    def test_receive_gpg_sign(self, tmp_files_path, receive_repo, source_repo,
                              source_server, gpg_homedir, monkeypatch):
        random_commit(source_repo, tmp_files_path, 'ref1')

        # Specifying a bogus GPG key should fail
        receive_repo.config.gpg_sign = ['DEADBEEF']
        receive_repo.config.gpg_homedir = str(gpg_homedir)
        with pytest.raises(GLib.Error) as excinfo:
            receive_repo.receive(['ref1'])
        assert excinfo.value.matches(Gio.io_error_quark(),
                                     Gio.IOErrorEnum.FAILED)

        # Specifying both key and homedir
        receive_repo.config.gpg_sign = [PGP_KEY_ID]
        receive_repo.config.gpg_homedir = str(gpg_homedir)
        merged = receive_repo.receive(['ref1'])
        assert merged == {'ref1'}
        refs = local_refs(receive_repo)
        assert refs.keys() == {'ref1'}

        # Validate the signature and make sure it was signed by the correct
        # key
        commit = refs['ref1']
        keyring_file = Gio.File.new_for_path(str(PGP_PUB_KEYRING))
        result = receive_repo.verify_commit_ext(commit, None, keyring_file)
        OSTree.GpgVerifyResult.require_valid_signature(result)
        assert OSTree.GpgVerifyResult.count_all(result) == 1
        sig = OSTree.GpgVerifyResult.get_all(result, 0).unpack()
        key_id = sig[OSTree.GpgSignatureAttr.FINGERPRINT]
        assert key_id == PGP_KEY_ID

        # Using the default homedir via GNUPGHOME
        monkeypatch.setenv('GNUPGHOME', str(gpg_homedir))
        receive_repo.config.gpg_sign = [PGP_KEY_ID]
        receive_repo.config.gpg_homedir = None
        wipe_repo(receive_repo)
        merged = receive_repo.receive(['ref1'])
        assert merged == {'ref1'}
        refs = local_refs(receive_repo)
        assert refs.keys() == {'ref1'}

    @needs_gpg
    def test_receive_gpg_verify(self, tmp_path, tmp_files_path, dest_repo,
                                source_repo, source_server, gpg_homedir,
                                monkeypatch):
        # Specifying a missing GPG keyring should fail
        keyring_path = str(tmp_path / 'missing.gpg')
        config = receive.OTReceiveConfig(gpg_verify=True,
                                         gpg_trustedkeys=keyring_path,
                                         update=False)
        repo_path = str(dest_repo.path)
        with pytest.raises(receive.OTReceiveConfigError) as excinfo:
            receive.OTReceiveRepo(repo_path, source_server.url, config)
        assert str(excinfo.value) == (
            f'gpg_trustedkeys keyring "{keyring_path}" does not exist'
        )

        # Receiving an unsigned commit should fail.
        random_commit(source_repo, tmp_files_path, 'ref1')
        config = receive.OTReceiveConfig(gpg_verify=True,
                                         gpg_trustedkeys=str(PGP_PUB_KEYRING),
                                         update=False)
        repo = receive.OTReceiveRepo(repo_path, source_server.url, config)
        with pytest.raises(GLib.Error) as excinfo:
            repo.receive(['ref1'])
        assert excinfo.value.matches(OSTree.gpg_error_quark(),
                                     OSTree.GpgError.NO_SIGNATURE)

        # Receiving a signed commit should succeed.
        random_commit(source_repo, tmp_files_path, 'ref1',
                      gpg_key_id=PGP_KEY_ID, gpg_homedir=str(gpg_homedir))
        config = receive.OTReceiveConfig(gpg_verify=True,
                                         gpg_trustedkeys=str(PGP_PUB_KEYRING),
                                         update=False)
        repo = receive.OTReceiveRepo(repo_path, source_server.url, config)
        wipe_repo(repo)
        merged = repo.receive(['ref1'])
        assert merged == {'ref1'}
        refs = local_refs(repo)
        assert refs.keys() == {'ref1'}

        # Using an ASCII armored key instead of a PGP keyring should
        # also work.
        random_commit(source_repo, tmp_files_path, 'ref1',
                      gpg_key_id=PGP_KEY_ID, gpg_homedir=str(gpg_homedir))
        config = receive.OTReceiveConfig(gpg_verify=True,
                                         gpg_trustedkeys=str(PGP_PUB),
                                         update=False)
        repo = receive.OTReceiveRepo(repo_path, source_server.url, config)
        wipe_repo(repo)
        merged = repo.receive(['ref1'])
        assert merged == {'ref1'}
        refs = local_refs(repo)
        assert refs.keys() == {'ref1'}

        # Using the user's default keyring.
        random_commit(source_repo, tmp_files_path, 'ref1',
                      gpg_key_id=PGP_KEY_ID, gpg_homedir=str(gpg_homedir))
        monkeypatch.setenv('XDG_CONFIG_HOME', str(tmp_path))
        keyring = tmp_path / 'ostree/ostree-receive-trustedkeys.gpg'
        keyring.parent.mkdir(exist_ok=True)
        keyring.symlink_to(PGP_PUB_KEYRING)
        config = receive.OTReceiveConfig(gpg_verify=True, update=False)
        repo = receive.OTReceiveRepo(repo_path, source_server.url, config)
        wipe_repo(repo)
        merged = repo.receive(['ref1'])
        assert merged == {'ref1'}
        refs = local_refs(repo)
        assert refs.keys() == {'ref1'}

    @needs_ed25519
    def test_receive_ed25519_sign(self, tmp_files_path, tmp_path, receive_repo,
                                  source_repo, ed25519_private_keyfile):
        random_commit(source_repo, tmp_files_path, 'ref1')

        # Specifying a missing keyfile should fail.
        keyfile_path = str(tmp_path / 'missing')
        receive_repo.config.sign_keyfiles = [keyfile_path]
        with pytest.raises(receive.OTReceiveConfigError,
                           match=f'sign_keyfiles keyfile "{keyfile_path}"'
                                 + ' does not exist'):
            receive_repo.receive(['ref1'])

        # Specifying the key.
        receive_repo.config.sign_keyfiles = [ed25519_private_keyfile]
        merged = receive_repo.receive(['ref1'])
        assert merged == {'ref1'}
        refs = local_refs(receive_repo)
        assert refs.keys() == {'ref1'}

        # Validate the signature and make sure it was signed by the correct
        # key.
        sign = get_ostree_ed25519_sign()
        sign.set_pk(GLib.Variant('s', ED25519_PUBLIC_KEY))
        commit = refs['ref1']
        assert sign.commit_verify(receive_repo, commit)

    @needs_ed25519
    def test_receive_ed25519_verify(self, tmp_path, tmp_files_path, dest_repo,
                                    source_repo, source_server,
                                    ed25519_public_keyfile, monkeypatch):
        # Specifying a missing keyfile should fail.
        keyfile_path = str(tmp_path / 'missing')
        config = receive.OTReceiveConfig(sign_verify=True,
                                         sign_trustedkeyfile=keyfile_path,
                                         update=False)
        repo_path = str(dest_repo.path)
        with pytest.raises(receive.OTReceiveConfigError,
                           match='sign_trustedkeyfile keyfile'
                                 + f' "{keyfile_path}" does not'
                                 + ' exist') as excinfo:
            receive.OTReceiveRepo(repo_path, source_server.url, config)

        # Receiving an unsigned commit should fail.
        random_commit(source_repo, tmp_files_path, 'ref1')
        config = receive.OTReceiveConfig(
            sign_verify=True,
            sign_trustedkeyfile=ed25519_public_keyfile,
            update=False)
        repo = receive.OTReceiveRepo(repo_path, source_server.url, config)
        with pytest.raises(GLib.Error, match="Can't verify commit") as excinfo:
            repo.receive(['ref1'])
        assert excinfo.value.matches(Gio.io_error_quark(),
                                     Gio.IOErrorEnum.FAILED)

        # Receiving a signed commit should succeed.
        random_commit(source_repo, tmp_files_path, 'ref1',
                      ed25519_key=ED25519_PRIVATE_KEY)
        config = receive.OTReceiveConfig(
            sign_verify=True,
            sign_trustedkeyfile=ed25519_public_keyfile,
            update=False)
        repo = receive.OTReceiveRepo(repo_path, source_server.url, config)
        wipe_repo(repo)
        merged = repo.receive(['ref1'])
        assert merged == {'ref1'}
        refs = local_refs(repo)
        assert refs.keys() == {'ref1'}

        # Using the user's default keyfile.
        random_commit(source_repo, tmp_files_path, 'ref1',
                      ed25519_key=ED25519_PRIVATE_KEY)
        monkeypatch.setenv('XDG_CONFIG_HOME', str(tmp_path))
        keyring = tmp_path / 'ostree/ostree-receive-trustedkeyfile.ed25519'
        keyring.parent.mkdir(exist_ok=True)
        keyring.symlink_to(ed25519_public_keyfile)
        config = receive.OTReceiveConfig(sign_verify=True, update=False)
        repo = receive.OTReceiveRepo(repo_path, source_server.url, config)
        wipe_repo(repo)
        merged = repo.receive(['ref1'])
        assert merged == {'ref1'}
        refs = local_refs(repo)
        assert refs.keys() == {'ref1'}

    @needs_ostree
    def test_update_repo_metadata(self, tmp_files_path, receive_repo):
        summary = Path(receive_repo.path) / 'summary'

        receive_repo.update_repo_metadata()
        assert summary.exists()
        summary_refs, summary_metadata = get_summary_variant(summary)
        ref_names = {ref[0] for ref in summary_refs}
        assert ref_names == {'ostree-metadata'}
        assert 'xa.cache' not in summary_metadata

        random_commit(receive_repo, tmp_files_path, 'someref')
        receive_repo.update_repo_metadata()
        assert summary.exists()
        summary_refs, summary_metadata = get_summary_variant(summary)
        ref_names = {ref[0] for ref in summary_refs}
        assert ref_names == {'ostree-metadata', 'someref'}
        assert 'xa.cache' not in summary_metadata

    @needs_flatpak
    def test_update_repo_metadata_flatpak(self, tmp_files_path, receive_repo):
        summary = Path(receive_repo.path) / 'summary'

        random_commit(receive_repo, tmp_files_path,
                      'app/com.example.App/x86_64/stable')
        receive_repo.update_repo_metadata()
        assert summary.exists()
        summary_refs, summary_metadata = get_summary_variant(summary)
        ref_names = {ref[0] for ref in summary_refs}

        # Flatpak < 1.10 creates the ostree-metadata commit when the
        # repo has a collecton ID, but newer versions don't. Add it to
        # the generated set if it's not there already so the expected
        # set is consistent.
        ref_names.add('ostree-metadata')
        assert ref_names == {
            'ostree-metadata',
            'app/com.example.App/x86_64/stable',
            'appstream/x86_64',
            'appstream2/x86_64',
        }
        assert 'xa.cache' in summary_metadata

    @needs_gpg
    @needs_ostree
    def test_update_repo_metadata_gpg_sign(self, receive_repo, gpg_homedir):
        receive_repo.config.gpg_sign = [PGP_KEY_ID]
        receive_repo.config.gpg_homedir = str(gpg_homedir)
        receive_repo.update_repo_metadata()

        summary = Path(receive_repo.path) / 'summary'
        summary_sig = summary.with_suffix('.sig')
        assert summary.exists()
        assert summary_sig.exists()

    @needs_ed25519
    @needs_ostree
    def test_update_repo_metadata_ed25519_sign(self, receive_repo,
                                               ed25519_private_keyfile):
        receive_repo.config.sign_keyfiles = [ed25519_private_keyfile]
        receive_repo.update_repo_metadata()

        summary = Path(receive_repo.path) / 'summary'
        summary_sig = summary.with_suffix('.sig')
        assert summary.exists()
        assert summary_sig.exists()

    def test_update_repo_hook(self, receive_repo, tmp_path, monkeypatch):
        dumpenv = os.path.join(TESTSDIR, 'dumpenv')
        dumpenv_dest = tmp_path / 'dumpenv.json'
        monkeypatch.setenv('DUMPENV_DEST', str(dumpenv_dest))

        # Exported environment variables
        receive_repo.config.update_hook = dumpenv
        receive_repo.update_repo_hook(['foo', 'bar'])
        with dumpenv_dest.open() as f:
            data = json.load(f)
        assert data['env']['OSTREE_RECEIVE_REPO'] == str(
            receive_repo.path.absolute()
        )
        assert data['env']['OSTREE_RECEIVE_REFS'] == 'foo bar'

        # Wrong refs passed
        with pytest.raises(TypeError):
            receive_repo.update_repo_hook(None)

        # No hook configured
        receive_repo.config.update_hook = None
        with pytest.raises(receive.OTReceiveConfigError) as excinfo:
            receive_repo.update_repo_hook([])
        assert str(excinfo.value) == 'update_hook not set in configuration'

        # Missing or non-executable hook
        hook = tmp_path / 'hook'
        receive_repo.config.update_hook = str(hook)
        with pytest.raises(FileNotFoundError):
            receive_repo.update_repo_hook([])
        hook.touch()
        with pytest.raises(PermissionError):
            receive_repo.update_repo_hook([])

        # Hook argument parsing
        receive_repo.config.update_hook = f'{dumpenv} foo bar'
        receive_repo.update_repo_hook([])
        with dumpenv_dest.open() as f:
            data = json.load(f)
        assert data['args'] == [dumpenv, 'foo', 'bar']

        receive_repo.config.update_hook = f'{dumpenv} "foo bar"'
        receive_repo.update_repo_hook([])
        with dumpenv_dest.open() as f:
            data = json.load(f)
        assert data['args'] == [dumpenv, 'foo bar']

        receive_repo.config.update_hook = fr'{dumpenv} foo\ bar'
        receive_repo.update_repo_hook([])
        with dumpenv_dest.open() as f:
            data = json.load(f)
        assert data['args'] == [dumpenv, 'foo bar']

    def test_receive_missing(self, tmp_files_path, receive_repo, source_repo,
                             source_server):
        random_commit(source_repo, tmp_files_path, 'ref1')

        with pytest.raises(receive.OTReceiveError) as excinfo:
            receive_repo.receive(['missing'])
        assert str(excinfo.value) == \
            'Could not find ref missing in summary file'

        with pytest.raises(receive.OTReceiveError) as excinfo:
            receive_repo.receive(['missing', 'ref1'])
        assert str(excinfo.value) == \
            'Could not find ref missing in summary file'

    def test_receive_specific(self, tmp_files_path, receive_repo, source_repo,
                              source_server):
        random_commit(source_repo, tmp_files_path, 'ref1')
        random_commit(source_repo, tmp_files_path, 'ref2')

        merged = receive_repo.receive(['ref1'])
        assert merged == {'ref1'}
        refs = local_refs(receive_repo)
        assert refs.keys() == {'ref1'}

        merged = receive_repo.receive(['ref1', 'ref2'])
        assert merged == {'ref2'}
        refs = local_refs(receive_repo)
        assert refs.keys() == {'ref1', 'ref2'}

        merged = receive_repo.receive(['ref1', 'ref2'])
        assert merged == set()
        refs = local_refs(receive_repo)
        assert refs.keys() == {'ref1', 'ref2'}

    def test_receive_all(self, tmp_files_path, receive_repo, source_repo,
                         source_server):
        random_commit(source_repo, tmp_files_path, 'ref1')
        random_commit(source_repo, tmp_files_path, 'ref2')
        random_commit(source_repo, tmp_files_path, 'appstream/x86_64')
        random_commit(source_repo, tmp_files_path, 'appstream2/x86_64')
        random_commit(source_repo, tmp_files_path, 'ostree-metadata')
        source_refs = local_refs(source_repo)
        assert source_refs.keys() == {
            'ref1',
            'ref2',
            'appstream/x86_64',
            'appstream2/x86_64',
            'ostree-metadata',
        }

        merged = receive_repo.receive([])
        assert merged == {'ref1', 'ref2'}
        refs = local_refs(receive_repo)
        assert refs.keys() == {'ref1', 'ref2'}

        merged = receive_repo.receive([])
        assert merged == set()
        refs = local_refs(receive_repo)
        assert refs.keys() == {'ref1', 'ref2'}

    def test_root(self, tmp_path, tmp_files_path, source_server):
        url = source_server.url
        root = tmp_path / 'pub/repos'
        root.mkdir(parents=True)
        config = receive.OTReceiveConfig(root=str(root), update=False)
        root_tmp_repo = TmpRepo(root / 'root-dest')
        non_root_tmp_repo = TmpRepo(tmp_path / 'non-root-dest')

        # Requesting a repo outside the root should fail
        repo_path = non_root_tmp_repo.path
        logger.debug('Repo path %s', repo_path)
        with pytest.raises(receive.OTReceiveError) as excinfo:
            receive.OTReceiveRepo(str(repo_path), url, config)
        assert str(excinfo.value) == (
            f'repo {non_root_tmp_repo.path} not found'
        )

        # Absolute path under the root should work
        repo_path = root_tmp_repo.path.resolve()
        assert repo_path.is_absolute()
        logger.debug('Repo path %s', repo_path)
        with receive.OTReceiveRepo(str(repo_path), url, config):
            pass

        # Relative path under the root should work
        repo_path = root_tmp_repo.path.relative_to(root)
        assert not repo_path.is_absolute()
        logger.debug('Repo path %s', repo_path)
        with receive.OTReceiveRepo(str(repo_path), url, config):
            pass


class TestConfig:
    """Tests for OTReceiveConfig"""
    def test_defaults(self):
        config = receive.OTReceiveConfig()
        assert dataclasses.asdict(config) == {
            'root': None,
            'gpg_sign': [],
            'gpg_homedir': None,
            'gpg_verify': False,
            'gpg_trustedkeys': None,
            'sign_type': 'ed25519',
            'sign_keyfiles': [],
            'sign_verify': False,
            'sign_trustedkeyfile': None,
            'update': True,
            'update_hook': None,
            'log_level': 'INFO',
            'force': False,
            'dry_run': False,
        }

    def test_invalid(self):
        with pytest.raises(receive.OTReceiveConfigError) as excinfo:
            receive.OTReceiveConfig(update=None)
        assert str(excinfo.value) == (
            "update must be an instance of <class 'bool'>, "
            "but found <class 'NoneType'>"
        )

        with pytest.raises(receive.OTReceiveConfigError) as excinfo:
            receive.OTReceiveConfig(update='true')
        assert str(excinfo.value) == (
            "update must be an instance of <class 'bool'>, "
            "but found <class 'str'>"
        )

    def test_default_paths(self, tmp_path, monkeypatch):
        assert receive.OTReceiveConfig.default_paths() == [
            Path('/etc/ostree/ostree-receive.conf'),
            Path('~/.config/ostree/ostree-receive.conf'),
        ]

        monkeypatch.setenv('XDG_CONFIG_HOME', str(tmp_path))
        assert receive.OTReceiveConfig.default_paths() == [
            Path('/etc/ostree/ostree-receive.conf'),
            tmp_path / 'ostree/ostree-receive.conf',
        ]

        monkeypatch.setenv('OSTREE_RECEIVE_CONF', str(tmp_path))
        assert receive.OTReceiveConfig.default_paths() == [tmp_path]

    def test_load_valid(self, tmp_path):
        path = tmp_path / 'ostree-receive.conf'
        data = {
            'root': str(tmp_path / 'pub/repos'),
            'gpg_sign': ['01234567', '89ABCDEF'],
            'gpg_homedir': str(tmp_path / 'gnupg'),
            'gpg_verify': True,
            'gpg_trustedkeys': str(tmp_path / 'trustedkeys.gpg'),
            'sign_type': 'ed25519',
            'sign_keyfiles': [
                str(tmp_path / 'signkey1'),
                str(tmp_path / 'signkey2'),
            ],
            'sign_verify': True,
            'sign_trustedkeyfile': str(tmp_path / 'trustedkey'),
            'update': False,
            'update_hook': '/foo/bar baz',
            'log_level': 'DEBUG',
            'force': True,
            'dry_run': True,
        }
        with path.open('w') as f:
            yaml.dump(data, f)

        config = receive.OTReceiveConfig.load([path])
        assert dataclasses.asdict(config) == data

    def test_load_none(self):
        config = receive.OTReceiveConfig.load([])
        assert config == receive.OTReceiveConfig()

    def test_load_empty(self, tmp_path, caplog):
        caplog.set_level(logging.DEBUG, receive.logger.name)
        path = tmp_path / 'ostree-receive.conf'
        path.touch()
        config = receive.OTReceiveConfig.load([path])
        assert config == receive.OTReceiveConfig()

        expected_log_record = (
            receive.logger.name,
            logging.DEBUG,
            f'Ignoring empty config file {path}'
        )
        assert expected_log_record in caplog.record_tuples

    def test_load_missing(self, tmp_path, caplog):
        caplog.set_level(logging.DEBUG, receive.logger.name)
        path = tmp_path / 'ostree-receive.conf'
        config = receive.OTReceiveConfig.load([path])
        assert config == receive.OTReceiveConfig()

        expected_log_record = (
            receive.logger.name,
            logging.DEBUG,
            f'Skipping missing config file {path}'
        )
        assert expected_log_record in caplog.record_tuples

    def test_load_multiple(self, tmp_path):
        path1 = tmp_path / 'receive1.conf'
        data = {
            'log_level': 'DEBUG',
        }
        with path1.open('w') as f:
            yaml.dump(data, f)

        path2 = tmp_path / 'receive2.conf'
        data = {
            'log_level': 'WARNING',
        }
        with path2.open('w') as f:
            yaml.dump(data, f)

        config = receive.OTReceiveConfig.load([path1, path2])
        assert config.log_level == 'WARNING'

    def test_load_unknown(self, tmp_path, caplog):
        caplog.set_level(logging.WARNING, receive.logger.name)
        path = tmp_path / 'ostree-receive.conf'
        data = {
            'fake_option': False,
        }
        with path.open('w') as f:
            yaml.dump(data, f)

        config = receive.OTReceiveConfig.load([path])
        assert config == receive.OTReceiveConfig()

        expected_log_record = (
            receive.logger.name,
            logging.WARNING,
            f'Unrecognized option fake_option in config file {path}',
        )
        assert expected_log_record in caplog.record_tuples

    def test_load_invalid(self, tmp_path):
        with pytest.raises(receive.OTReceiveConfigError) as excinfo:
            receive.OTReceiveConfig.load([True])
        assert str(excinfo.value) == (
            'expected str, bytes or os.PathLike object, not bool'
        )

        path = tmp_path / 'ostree-receive.conf'
        data = {
            'update': None,
        }
        with path.open('w') as f:
            yaml.dump(data, f)

        with pytest.raises(receive.OTReceiveConfigError) as excinfo:
            receive.OTReceiveConfig.load([path])
        assert str(excinfo.value) == (
            "update must be an instance of <class 'bool'>, "
            "but found <class 'NoneType'>"
        )

        data = ['not', 'a', 'mapping']
        with path.open('w') as f:
            yaml.dump(data, f)

        with pytest.raises(receive.OTReceiveConfigError) as excinfo:
            receive.OTReceiveConfig.load([path])
        assert str(excinfo.value) == (
            f'Config file {path} is not a YAML mapping'
        )

    def test_load_env(self, tmp_path, monkeypatch):
        path = tmp_path / 'ostree-receive.conf'
        data = {
            'log_level': 'DEBUG',
        }
        with path.open('w') as f:
            yaml.dump(data, f)

        monkeypatch.setenv('OSTREE_RECEIVE_CONF', str(path))
        config = receive.OTReceiveConfig.load()
        assert config.log_level == 'DEBUG'

    def test_load_args(self, caplog):
        caplog.set_level(logging.DEBUG, receive.logger.name)
        ap = argparse.ArgumentParser()
        ap.add_argument('--log-level', default='WARNING')
        ap.add_argument('--someopt', default='someval')
        ap.add_argument('pos')
        args = ap.parse_args(['foo'])

        config = receive.OTReceiveConfig.load(paths=[], args=args)
        assert dataclasses.asdict(config) == {
            'root': None,
            'gpg_sign': [],
            'gpg_homedir': None,
            'gpg_verify': False,
            'gpg_trustedkeys': None,
            'sign_type': 'ed25519',
            'sign_keyfiles': [],
            'sign_verify': False,
            'sign_trustedkeyfile': None,
            'update': True,
            'update_hook': None,
            'log_level': 'WARNING',
            'force': False,
            'dry_run': False,
        }

        expected_log_record = (
            receive.logger.name,
            logging.DEBUG,
            'Ignoring argument someopt'
        )
        assert expected_log_record in caplog.record_tuples

        expected_log_record = (
            receive.logger.name,
            logging.DEBUG,
            'Ignoring argument pos'
        )
        assert expected_log_record in caplog.record_tuples

    def test_load_args_invalid(self):
        with pytest.raises(receive.OTReceiveConfigError) as excinfo:
            receive.OTReceiveConfig.load(paths=[], args='foo')
        assert str(excinfo.value) == (
            'args is not an argparse.Namespace instance'
        )

    def test_load_conf_and_args(self, tmp_path):
        path = tmp_path / 'ostree-receive.conf'
        data = {
            'log-level': 'DEBUG',
        }
        with path.open('w') as f:
            yaml.dump(data, f)

        ap = argparse.ArgumentParser()
        ap.add_argument('--log-level', default='WARNING')
        args = ap.parse_args([])

        config = receive.OTReceiveConfig.load(paths=[path], args=args)
        assert config.log_level == 'WARNING'


class TestArgParser:
    def test_no_repo(self, capsys):
        ap = receive.OTReceiveArgParser()
        with pytest.raises(SystemExit) as excinfo:
            ap.parse_args([])
        assert excinfo.value.code == 2
        out, err = capsys.readouterr()
        assert out == ''
        assert err.endswith(
            'error: the following arguments are required: REPO, URL\n'
        )

    def test_no_url(self, capsys):
        ap = receive.OTReceiveArgParser()
        with pytest.raises(SystemExit) as excinfo:
            ap.parse_args(['repo'])
        assert excinfo.value.code == 2
        out, err = capsys.readouterr()
        assert out == ''
        assert err.endswith(
            'error: the following arguments are required: URL\n'
        )

    def test_defaults(self):
        ap = receive.OTReceiveArgParser()
        args = ap.parse_args(['repo', 'url'])
        assert args == argparse.Namespace(
            repo='repo',
            url='url',
            refs=[],
        )

    def test_refs(self):
        ap = receive.OTReceiveArgParser()
        args = ap.parse_args(['repo', 'url', 'foo'])
        assert args.refs == ['foo']
        args = ap.parse_args(['repo', 'url', 'foo', 'bar', 'baz'])
        assert args.refs == ['foo', 'bar', 'baz']

    def test_update(self):
        ap = receive.OTReceiveArgParser()
        args = ap.parse_args(['--no-update', 'repo', 'url'])
        assert args.update is False

    def test_dry_run(self):
        ap = receive.OTReceiveArgParser()
        args = ap.parse_args(['-n', 'repo', 'url'])
        assert args.dry_run is True
        args = ap.parse_args(['--dry-run', 'repo', 'url'])
        assert args.dry_run is True

    def test_force(self):
        ap = receive.OTReceiveArgParser()
        args = ap.parse_args(['-f', 'repo', 'url'])
        assert args.force is True
        args = ap.parse_args(['--force', 'repo', 'url'])
        assert args.force is True

    def test_log_level(self):
        ap = receive.OTReceiveArgParser()
        args = ap.parse_args(['-v', 'repo', 'url'])
        assert args.log_level == 'DEBUG'
        args = ap.parse_args(['--verbose', 'repo', 'url'])
        assert args.log_level == 'DEBUG'
        args = ap.parse_args(['-q', 'repo', 'url'])
        assert args.log_level == 'WARNING'
        args = ap.parse_args(['--quiet', 'repo', 'url'])
        assert args.log_level == 'WARNING'
